import 'dart:async';
import 'package:flutter/material.dart';
import 'package:zwap_design_system/atoms/colors/zwapColors.dart';
import 'package:zwap_design_system/atoms/constants/zwapConstants.dart';
import 'package:zwap_design_system/atoms/layouts/infiniteScroll/zwapInfiniteScroll.dart';
import 'package:zwap_design_system/atoms/text/base/zwapText.dart';
import 'package:zwap_design_system/atoms/typography/zwapTypography.dart';
import 'package:zwap_design_system/modelData/api.dart';

enum SuggestionState {
  /// always show suggestions
  enabled,

  /// hide the suggestions on initial focus
  hidden,

  /// show suggestions only on tap
  onTap,
}

// enum to define the Focus of the searchfield when a suggestion is tapped
enum SuggestionAction {
  /// shift to next focus
  next,

  /// close keyboard and unfocus
  unfocus,
}

class ZwapSearchField<T extends Object> extends StatefulWidget {
  /// Data source to perform search.
  final List<String> suggestions;

  /// Callback to return the selected suggestion.
  final Function(String?)? onTap;

  /// Hint for the [ZwapSearchField].
  final String? hint;

  /// Define a [TextInputAction] that is called when the field is submitted
  final TextInputAction? textInputAction;

  /// The initial value to be selected for [ZwapSearchField]. The value
  /// must be present in [suggestions].
  ///
  /// When not specified, [hint] is shown instead of `initialValue`.
  final String? initialValue;

  /// Specifies [TextStyle] for search input.
  final TextStyle? searchStyle;

  /// Specifies [TextStyle] for suggestions.
  final TextStyle? suggestionStyle;

  /// Specifies [InputDecoration] for search input [TextField].
  ///
  /// When not specified, the default value is [InputDecoration] initialized
  /// with [hint].
  final InputDecoration? searchInputDecoration;

  /// defaults to SuggestionState.hidden
  final SuggestionState suggestionState;

  /// Specifies the [SuggestionAction] called on suggestion tap.
  final SuggestionAction? suggestionAction;

  final PageData<T> Function(String query, int pageNumber) fetchMoreData;

  /// Specifies [BoxDecoration] for suggestion list. The property can be used to add [BoxShadow],
  /// and much more. For more information, checkout [BoxDecoration].
  ///
  /// Default value,
  ///
  /// ```dart
  /// BoxDecoration(
  ///   color: Theme.of(context).colorScheme.surface,
  ///   boxShadow: [
  ///     BoxShadow(
  ///       color: onSurfaceColor.withOpacity(0.1),
  ///       blurRadius: 8.0, // soften the shadow
  ///       spreadRadius: 2.0, //extend the shadow
  ///       offset: Offset(
  ///         2.0,
  ///         5.0,
  ///       ),
  ///     ),
  ///   ],
  /// )
  /// ```
  final BoxDecoration? suggestionsDecoration;

  /// Specifies [BoxDecoration] for items in suggestion list. The property can be used to add [BoxShadow],
  /// and much more. For more information, checkout [BoxDecoration].
  ///
  /// Default value,
  ///
  /// ```dart
  /// BoxDecoration(
  ///   border: Border(
  ///     bottom: BorderSide(
  ///       color: widget.marginColor ??
  ///         onSurfaceColor.withOpacity(0.1),
  ///     ),
  ///   ),
  /// )
  final BoxDecoration? suggestionItemDecoration;

  /// Specifies height for item suggestion.
  ///
  /// When not specified, the default value is `35.0`.
  final double itemHeight;

  /// Specifies the color of margin between items in suggestions list.
  ///
  /// When not specified, the default value is `Theme.of(context).colorScheme.onSurface.withOpacity(0.1)`.
  final Color? marginColor;

  /// Specifies the number of suggestions that can be shown in viewport.
  ///
  /// When not specified, the default value is `5`.
  /// if the number of suggestions is less than 5, then [maxSuggestionsInViewPort]
  /// will be the length of [suggestions]
  final int maxSuggestionsInViewPort;

  /// Specifies the `TextEditingController` for [ZwapSearchField].
  final TextEditingController? controller;

  /// `validator` for the [ZwapSearchField]
  /// to make use of this validator, The
  /// SearchField widget needs to be wrapped in a Form
  /// and pass it a Global key
  /// and write your validation logic in the validator
  /// you can define a global key
  ///
  ///  ```
  ///  Form(
  ///   key: _formKey,
  ///   child: SearchField(
  ///     suggestions: _statesOfIndia,
  ///     validator: (state) {
  ///       if (!_statesOfIndia.contains(state) || state.isEmpty) {
  ///         return 'Please Enter a valid State';
  ///       }
  ///       return null;
  ///     },
  ///   )
  /// ```
  /// You can then validate the form by calling
  /// the validate function of the form
  ///
  /// `_formKey.currentState.validate();`
  ///
  ///
  ///
  final String? Function(String?)? validator;

  /// if false the suggestions will be shown below
  /// the searchfield along the Y-axis.
  /// if true the suggestions will be shown floating like the
  /// along the Z-axis
  /// defaults to ```true```
  final bool hasOverlay;

  ZwapSearchField({
    Key? key,
    required this.suggestions,
    required this.fetchMoreData,
    this.initialValue,
    this.hint,
    this.hasOverlay = true,
    this.searchStyle,
    this.marginColor,
    this.controller,
    this.validator,
    this.suggestionState = SuggestionState.hidden,
    this.itemHeight = 35.0,
    this.suggestionsDecoration,
    this.suggestionStyle,
    this.searchInputDecoration,
    this.suggestionItemDecoration,
    this.maxSuggestionsInViewPort = 5,
    this.onTap,
    this.textInputAction,
    this.suggestionAction,
  })  : assert(
  (initialValue != null && suggestions.contains(initialValue)) ||
      initialValue == null,
  'Initial value should either be null or should be present in suggestions list.'),
        super(key: key);

  @override
  _ZwapSearchFieldState<T> createState() => _ZwapSearchFieldState<T>();
}

class _ZwapSearchFieldState<T extends Object> extends State<ZwapSearchField>{
  final StreamController<List<String?>?> sourceStream =
  StreamController<List<String>?>.broadcast();
  final FocusNode _focus = FocusNode();
  bool sourceFocused = false;
  TextEditingController? sourceController;

  @override
  void dispose() {
    _focus.dispose();
    sourceStream.close();
    super.dispose();
  }

  void initialize() {
    _focus.addListener(() {
      setState(() {
        sourceFocused = _focus.hasFocus;
      });
      if (widget.hasOverlay) {
        if (sourceFocused) {
          if (widget.initialValue == null) {
            if (widget.suggestionState == SuggestionState.enabled) {
              Future.delayed(Duration(milliseconds: 100), () {
                sourceStream.sink.add(widget.suggestions);
              });
            }
          }
          _overlayEntry = _createOverlay();
          Overlay.of(context)!.insert(_overlayEntry);
        } else {
          _overlayEntry.remove();
        }
      } else if (sourceFocused) {
        if (widget.initialValue == null) {
          if (widget.suggestionState == SuggestionState.enabled) {
            Future.delayed(Duration(milliseconds: 100), () {
              sourceStream.sink.add(widget.suggestions);
            });
          }
        }
      }
    });
  }

  late OverlayEntry _overlayEntry;
  @override
  void initState() {
    super.initState();
    sourceController = widget.controller ?? TextEditingController();
    initialize();
    WidgetsBinding.instance!.addPostFrameCallback((_) {
      if (widget.initialValue == null || widget.initialValue!.isEmpty) {

      } else {
        sourceController!.text = widget.initialValue!;
        sourceStream.sink.add([widget.initialValue]);
      }
    });
  }

  @override
  void didUpdateWidget(covariant ZwapSearchField oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.controller != widget.controller) {
      sourceController = widget.controller ?? TextEditingController();
    }
    if (oldWidget.hasOverlay != widget.hasOverlay) {
      if (widget.hasOverlay) {
        initialize();
      } else {
        if (_overlayEntry.mounted) {
          _overlayEntry.remove();
        }
      }
      setState(() {});
    }
  }

  Widget _suggestionElement(String suggestionText, bool borderBoolean){
    final onSurfaceColor = Theme.of(context).colorScheme.onSurface;
    return GestureDetector(
      onTap: () {
        sourceController!.text = suggestionText;
        sourceController!.selection = TextSelection.fromPosition(
          TextPosition(
            offset: sourceController!.text.length,
          ),
        );

        // suggestion action
        if (widget.suggestionAction != null) {
          if (widget.suggestionAction == SuggestionAction.next) {
            _focus.nextFocus();
          } else if (widget.suggestionAction ==
              SuggestionAction.unfocus) {
            _focus.unfocus();
          }
        }

        // hide the suggestions
        sourceStream.sink.add(null);
        if (widget.onTap != null) {
          widget.onTap!(suggestionText);
        }
      },
      child: Container(
        height: widget.itemHeight,
        padding: EdgeInsets.symmetric(horizontal: 5) +
            EdgeInsets.only(left: 8),
        width: double.infinity,
        alignment: Alignment.centerLeft,
        decoration: widget.suggestionItemDecoration?.copyWith(
          border: Border(
            bottom: BorderSide(
              color: widget.marginColor ??
                  onSurfaceColor.withOpacity(0.1),
            ),
          ),
        ) ??
            BoxDecoration(
              border: borderBoolean
                  ? null
                  : Border(
                bottom: BorderSide(
                  color: widget.marginColor ??
                      onSurfaceColor.withOpacity(0.1),
                ),
              ),
            ),
        child: Text(
          suggestionText,
          style: widget.suggestionStyle,
        ),
      ),
    );
  }

  Future<PageData<String>> fetchData(int pageNumber, String? item) async{
    PageData<T> queryData = await widget.fetchMoreData(item ?? sourceController!.text, pageNumber);

    this.sourceStream.sink.add(roles);
    return PageData<String>(
        data: roles,
        count: roles.length,
        next: queryData.next,
        previous: queryData.previous
    );
  }

  Widget _suggestionsBuilder() {
    final onSurfaceColor = Theme.of(context).colorScheme.onSurface;
    return StreamBuilder<List<String?>?>(
      stream: sourceStream.stream,
      builder: (BuildContext context, AsyncSnapshot<List<String?>?> snapshot) {
        if (snapshot.data == null || snapshot.data!.isEmpty || !sourceFocused) {
          return Container();
        } else {
          if (snapshot.data!.length > widget.maxSuggestionsInViewPort) {
            height = widget.itemHeight * widget.maxSuggestionsInViewPort;
          } else if (snapshot.data!.length == 1) {
            height = widget.itemHeight;
          } else {
            height = snapshot.data!.length * widget.itemHeight;
          }
          Widget lastWidget =ZwapInfiniteScroll<String>(
              fetchMoreData: (int pageNumber) => this.fetchData(pageNumber, null),
              getChildWidget: (String role) => this._suggestionElement(role, true),
              zwapInfiniteScrollType: ZwapInfiniteScrollType.listView,
              mainSizeDirection: 300
          );
          Widget listWidget = ListView.builder(
            reverse: isUp,
            padding: EdgeInsets.zero,
            itemCount: snapshot.data!.length,
            physics: snapshot.data!.length == 1
                ? NeverScrollableScrollPhysics()
                : ScrollPhysics(),
            itemBuilder: (context, index) => this._suggestionElement(snapshot.data![index]!, index == snapshot.data!.length - 1),
          );
          return AnimatedContainer(
            duration: isUp ? Duration.zero : Duration(milliseconds: 300),
            height: height,
            alignment: Alignment.centerLeft,
            decoration: widget.suggestionsDecoration ??
                BoxDecoration(
                  color: Theme.of(context).colorScheme.surface,
                  boxShadow: [
                    BoxShadow(
                      color: onSurfaceColor.withOpacity(0.1),
                      blurRadius: 8.0, // soften the shadow
                      spreadRadius: 2.0, // extend the shadow
                      offset: widget.hasOverlay
                          ? Offset(
                        2.0,
                        5.0,
                      )
                          : Offset(1.0, 0.5),
                    ),
                  ],
                ),
            child: lastWidget,
          );
        }
      },
    );
  }

  Offset getYOffset(Offset widgetOffset, int resultCount) {
    final size = MediaQuery.of(context).size;
    final position = widgetOffset.dy;
    if ((position + height) < (size.height - widget.itemHeight * 2)) {
      return Offset(0, widget.itemHeight + 10.0);
    } else {
      if (resultCount > widget.maxSuggestionsInViewPort) {
        isUp = false;
        return Offset(
            0, -(widget.itemHeight * widget.maxSuggestionsInViewPort));
      } else {
        isUp = true;
        return Offset(0, -(widget.itemHeight * resultCount));
      }
    }
  }

  OverlayEntry _createOverlay() {
    final renderBox = context.findRenderObject() as RenderBox;
    final size = renderBox.size;
    final offset = renderBox.localToGlobal(Offset.zero);
    return OverlayEntry(
        builder: (context) => StreamBuilder<List<String?>?>(
            stream: sourceStream.stream,
            builder:
                (BuildContext context, AsyncSnapshot<List<String?>?> snapshot) {
              late var count = widget.maxSuggestionsInViewPort;
              if (snapshot.data != null) {
                count = snapshot.data!.length;
              }
              return Positioned(
                left: offset.dx,
                width: size.width,
                child: CompositedTransformFollower(
                    offset: getYOffset(offset, count),
                    link: _layerLink,
                    child: Material(child: _suggestionsBuilder())),
              );
            }));
  }

  /// It retrieves the borderColor in base of the current status
  InputBorder _getInputBorder() {
    OutlineInputBorder inputBorder = OutlineInputBorder(
        borderRadius: BorderRadius.all(Radius.circular(ZwapRadius.buttonRadius)),
        borderSide: BorderSide(
            color: ZwapColors.neutral300,
            width: 1,
            style: BorderStyle.solid
        )
    );
    return inputBorder;
  }

  /// It retrieves the borderColor in base of the current status
  InputBorder _getFocusedBorder() {
    OutlineInputBorder inputBorder = OutlineInputBorder(
        borderRadius: BorderRadius.all(Radius.circular(ZwapRadius.buttonRadius)),
        borderSide: BorderSide(
            color: ZwapColors.primary300,
            width: 1,
            style: BorderStyle.solid
        )
    );
    return inputBorder;
  }

  /// The field decoration for any text field
  InputDecoration _getTextFieldDecoration() {
    return new InputDecoration(
      enabledBorder: this._getInputBorder(),
      focusedBorder: this._getFocusedBorder(),
      hoverColor: ZwapColors.neutral300,
      hintText: widget.hint,
      hintStyle: getTextStyle(ZwapTextType.bodyRegular).apply(color: ZwapColors.neutral400),
    );
  }

  final LayerLink _layerLink = LayerLink();
  late double height;
  bool isUp = false;
  @override
  Widget build(BuildContext context) {
    height = 35*5;
    return Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        Padding(
          padding: EdgeInsets.only(bottom: 2.5),
          child: ZwapText(
            text: "Prova",
            zwapTextType: ZwapTextType.bodySemiBold,
            textColor: ZwapColors.neutral600,
          ),
        ),
        Padding(
          padding: EdgeInsets.only(top: 2.5),
          child: Column(
            children: [
              CompositedTransformTarget(
                link: _layerLink,
                child: TextFormField(
                  onTap: () {
                    /// only call that if [SuggestionState.onTap] is selected
                    if (!sourceFocused &&
                        widget.suggestionState == SuggestionState.onTap) {
                      setState(() {
                        sourceFocused = true;
                      });
                      Future.delayed(Duration(milliseconds: 100), () {
                        sourceStream.sink.add(widget.suggestions);
                      });
                    }
                  },
                  controller: widget.controller ?? sourceController,
                  focusNode: _focus,
                  validator: widget.validator,
                  style: widget.searchStyle,
                  textInputAction: widget.textInputAction,
                  decoration: _getTextFieldDecoration(),
                  onChanged: (item) async => await this.fetchData(1, item),
                ),
              ),
              if (!widget.hasOverlay)
                SizedBox(
                  height: 2,
                ),
              if (!widget.hasOverlay) _suggestionsBuilder()
            ],
          ),
        )
      ],
    );
  }
}